/* add dollar sign in console code-block */
div.highlight-console pre span.go::before {
  content: "$";
  margin-right: 10px;
  margin-left: 5px;
}

/* Create a custom api admonition */
div.admonition.admonition-api > .admonition-title::after {
  content: "\f121"; /* the fa-code icon */
}

/* increase width of the main content */
.bd-main .bd-content .bd-article-container {
  max-width: 100%; /* default is 60em */
}

/* increase width of the page */
.bd-page-width {
  max-width: 110rem; /* default is 88rem */
}
